import multiprocessing as mp
import os
import pytest
import random
import shutil
import sys

import pandas as pd

import tuneconfig
from tuneconfig.experiment import Experiment, ExperimentMode

sys.path.insert(0, os.path.abspath("tests"))
import conftest


@pytest.fixture(scope="function")
def experiment(config_factory):
    logdir = "/tmp/tuneconfig_4"
    experiment = Experiment(config_factory, logdir)
    experiment.start()
    yield experiment
    shutil.rmtree(logdir)


def test_run(experiment):
    num_samples = num_workers = 5

    results = experiment.run(conftest.exec_func, num_samples, num_workers)
    assert len(results) == len(experiment.config_iterator)

    for trial_id, trial_results in results.items():
        assert len(trial_results) == num_samples

        for pname, config, run_dir, run_id in trial_results:
            assert run_dir not in config
            assert os.path.exists(run_dir)


def test_run_in_append_mode(experiment):
    num_samples = num_workers = 5
    n_experiments = 3

    for _ in range(n_experiments):
        results = experiment.run(conftest.exec_func,
                                 num_samples, num_workers,
                                 mode=ExperimentMode.APPEND)
        assert len(results) == len(experiment.config_iterator)

    for config in experiment.config_iterator:
        _, trial_dir = experiment._get_trial(config)
        runs = experiment.get_run_dirs(trial_dir)
        assert len(runs) == num_samples * n_experiments


def test_run_in_skip_mode(experiment):
    num_samples = num_workers = 5

    experiment.run(conftest.exec_func, num_samples, num_workers)
    mtimes1 = {}
    for config in experiment.config_iterator:
        _, trial_dir = experiment._get_trial(config)
        runs = experiment.get_run_dirs(trial_dir)
        mtimes1[trial_dir] = [os.stat(path).st_mtime for path in runs]

    n_experiments = 3
    for _ in range(n_experiments):
        results = experiment.run(
            conftest.exec_func, num_samples, num_workers,
            mode=ExperimentMode.SKIP)
        assert len(results) == 0

        for config in experiment.config_iterator:
            _, trial_dir = experiment._get_trial(config)
            runs = experiment.get_run_dirs(trial_dir)
            assert mtimes1[trial_dir] == [
                os.stat(path).st_mtime for path in runs]

    for config in experiment.config_iterator:
        _, trial_dir = experiment._get_trial(config)
        runs = experiment.get_run_dirs(trial_dir)
        assert len(runs) == num_samples


def test_run_in_overwrite_mode(experiment):
    num_samples = num_workers = 3

    experiment.run(conftest.exec_func, num_samples, num_workers)
    mtimes1 = {}
    for config in experiment.config_iterator:
        _, trial_dir = experiment._get_trial(config)
        runs = experiment.get_run_dirs(trial_dir)
        mtimes1[trial_dir] = [os.stat(path).st_mtime for path in runs]

    experiment.run(
        conftest.exec_func, num_samples, num_workers,
        mode=ExperimentMode.OVERWRITE)

    mtimes2 = {}
    for config in experiment.config_iterator:
        _, trial_dir = experiment._get_trial(config)
        runs = experiment.get_run_dirs(trial_dir)
        mtimes2[trial_dir] = [os.stat(path).st_mtime for path in runs]

    for config in experiment.config_iterator:
        _, trial_dir = experiment._get_trial(config)
        for mtime1, mtime2 in zip(mtimes1[trial_dir], mtimes2[trial_dir]):
            assert mtime1 != mtime2



import os
from collections import namedtuple
import itertools
import random
import shutil

import multiprocessing as mp
import pandas as pd
import pytest

from tuneconfig import *


PlotConfig = namedtuple("PlotConfig", "targets x_axis y_axis anchors")


def exec_func(config):
    logdir = config["logdir"]
    run_id = logdir.split("/")[-1]
    del config["logdir"]
    os.makedirs(logdir)

    df = pd.DataFrame(
        {
            "foo": pd.Series([random.gauss(10.0, 1.0) for _ in range(10)]),
            "baz": pd.Series([random.gauss(20.0, 2.0) for _ in range(10)]),
            "bar": pd.Series([random.gauss(30.0, 3.0) for _ in range(10)]),
        }
    )
    filepath = os.path.join(logdir, "data.csv")
    df.to_csv(filepath, index=False)

    df = pd.DataFrame(
        {"test": pd.Series([random.randrange(-10.0, 10.0) for _ in range(10)]),}
    )
    filepath = os.path.join(logdir, "metric.csv")
    df.to_csv(filepath, index=False)

    return (mp.current_process().pid, config, logdir, run_id)


@pytest.fixture(scope="session")
def config_factory():
    def format_fn(param):
        fmt = {
            # "batch_size": "batch",
            # "horizon": "hr",
            # "learning_rate": "lr",
            # "optimizer": "opt",
            "num_samples": None
        }
        return fmt.get(param, param)

    return ConfigFactory(
        {
            "batch_size": grid_search([32, 64, 128]),
            "horizon": 40,
            "learning_rate": grid_search([0.01, 0.1]),
            "epochs": 1000,
            "optimizer": grid_search(["Adam", "RMSProp", "GradientDescent"]),
            "num_samples": 10,
        },
        format_fn=format_fn,
    )


@pytest.fixture(scope="session")
def experiment(config_factory):
    logdir = "/tmp/tuneconfig_1"
    experiment = Experiment(config_factory, logdir)
    experiment.start()
    yield experiment
    shutil.rmtree(logdir)


@pytest.fixture(scope="session")
def analysis(experiment):
    num_samples = num_workers = 10
    _ = experiment.run(exec_func, num_samples, num_workers)
    analysis = ExperimentAnalysis(experiment.logdir)
    analysis.setup()
    return analysis


@pytest.fixture(scope="session")
def trial(analysis):
    return analysis[0]


@pytest.fixture(scope="session")
def analysis_list(config_factory):
    basedir = "/tmp/tuneconfig_2"
    num_samples = num_workers = 3

    analyses = []
    for i in range(3):
        name = f"experiment{i}"
        logdir = os.path.join(basedir, name)
        experiment = Experiment(config_factory, logdir)
        experiment.start()
        _ = experiment.run(exec_func, num_samples, num_workers)
        exp_analysis = ExperimentAnalysis(experiment.logdir, name)
        exp_analysis.setup()
        analyses.append(exp_analysis)

    yield analyses

    for analysis in analyses:
        shutil.rmtree(analysis.logdir)


def confplot_idfn(fixture_value):
    return "_".join(filter(None, fixture_value))


@pytest.fixture(
    scope="module",
    params=list(
        itertools.product(
            ["single_target", "multiple_targets"],
            ["single_plot", "multiple_plots"],
            ["x_axis", None],
            ["y_axis", None],
        )
    ),
    ids=confplot_idfn,
)
def confplot(request):
    target_type, plot_type, x_axis, y_axis = request.param

    if target_type == "single_target":
        targets = ["metric:test"]
    else:
        targets = ["data:foo", "data:bar"]

    anchors = ["learning_rate=0.1", "horizon=40", "epochs=1000"]

    if x_axis:
        x_axis = "optimizer"
        if plot_type == "single_plot":
            anchors.append("batch_size=32")

    if y_axis:
        y_axis = "batch_size"
        if plot_type == "single_plot":
            anchors.append("optimizer=Adam")

    return PlotConfig(targets, x_axis, y_axis, anchors)


@pytest.fixture(scope="session")
def cfg_single_target_no_axes():
    targets = ["metric:test"]
    x_axis, y_axis = None, None
    anchors = ["batch_size=32", "learning_rate=0.1", "optimizer=Adam"]
    return PlotConfig(targets, x_axis, y_axis, anchors)


@pytest.fixture(scope="session")
def cfg_multiple_targets_no_axes():
    targets = ["data:foo", "data:baz"]
    x_axis, y_axis = None, None
    anchors = ["batch_size=32", "learning_rate=0.1", "optimizer=Adam"]
    return PlotConfig(targets, x_axis, y_axis, anchors)


@pytest.fixture(scope="session")
def cfg_single_target_multiple_plots_no_axes():
    targets = ["metric:test"]
    x_axis, y_axis = None, None
    anchors = ["batch_size=32", "learning_rate=0.1"]
    return PlotConfig(targets, x_axis, y_axis, anchors)


@pytest.fixture(scope="session")
def cfg_multiple_targets_multiple_plots_no_axes():
    targets = ["data:foo", "data:baz"]
    x_axis, y_axis = None, None
    anchors = ["batch_size=32", "learning_rate=0.1"]
    return PlotConfig(targets, x_axis, y_axis, anchors)


@pytest.fixture(scope="session")
def cfg_single_target_x_axis():
    targets = ["metric:test"]
    x_axis, y_axis = "optimizer", None
    anchors = ["batch_size=32", "learning_rate=0.1"]
    return PlotConfig(targets, x_axis, y_axis, anchors)


@pytest.fixture(scope="session")
def cfg_multiple_targets_x_axis():
    targets = ["data:foo", "data:baz"]
    x_axis, y_axis = "optimizer", None
    anchors = ["batch_size=32", "learning_rate=0.1"]
    return PlotConfig(targets, x_axis, y_axis, anchors)


@pytest.fixture(scope="session")
def cfg_single_target_y_axis():
    targets = ["metric:test"]
    x_axis, y_axis = None, "batch_size"
    anchors = ["learning_rate=0.1", "optimizer=Adam"]
    return PlotConfig(targets, x_axis, y_axis, anchors)


@pytest.fixture(scope="session")
def cfg_multiple_targets_y_axis():
    targets = ["data:foo", "data:baz"]
    x_axis, y_axis = None, "batch_size"
    anchors = ["learning_rate=0.1", "optimizer=Adam"]
    return PlotConfig(targets, x_axis, y_axis, anchors)


@pytest.fixture(scope="session")
def cfg_single_target_both_axis():
    targets = ["metric:test"]
    x_axis, y_axis = "optimizer", "learning_rate"
    anchors = ["batch_size=128"]
    return PlotConfig(targets, x_axis, y_axis, anchors)


@pytest.fixture(scope="session")
def cfg_multiple_targets_both_axis():
    targets = ["data:foo", "data:baz"]
    x_axis, y_axis = "optimizer", "learning_rate"
    anchors = ["batch_size=128"]
    return PlotConfig(targets, x_axis, y_axis, anchors)



from tuneconfig.config_factory import ConfigFactory, grid_search
from tuneconfig.experiment import Experiment
from tuneconfig.analysis import ExperimentAnalysis
from tuneconfig.plotter import ExperimentPlotter
from tuneconfig.launcher import run_experiment



from pprint import pprint
import time

import matplotlib.pyplot as plt
import pytest

from tuneconfig.plotter import ExperimentPlotter


@pytest.fixture
def plotter(analysis_list):
    kwargs = {
        "plot_type": "line",
        "figsize": (7, 5),
        "sharex": False,
        "sharey": False,
        "target_x_axis_label": None,
        "target_y_axis_label": None,
    }

    return ExperimentPlotter(*analysis_list, **kwargs)


def test_line_plot(plotter, confplot):
    fig = plotter.plot(*confplot)
    assert isinstance(fig, plt.Figure)
    # plt.show()
    fig.clear()
    plt.close(fig)


def test_bar_plot(plotter, confplot):
    plotter.kwargs["plot_type"] = "bar"
    targets, x_axis, y_axis, anchors = confplot
    targets = [f"mean/{target}" for target in targets]
    fig = plotter.plot(targets, x_axis, y_axis, anchors)
    assert isinstance(fig, plt.Figure)
    # plt.show()
    fig.clear()
    plt.close(fig)



from collections import defaultdict
import json
import os

import pandas as pd

from tuneconfig.experiment import Experiment
from tuneconfig.trial import Trial


class ExperimentAnalysis:
    """ExperimentAnalysis

    """

    RESULT_METRIC_SEPARATOR = ":"
    TRANSFORM_TARGET_SEPARATOR = "/"

    @classmethod
    def split_target(cls, target):
        transform = None
        if cls.TRANSFORM_TARGET_SEPARATOR in target:
            transform, target = target.split(cls.TRANSFORM_TARGET_SEPARATOR)
        result, metric = target.split(cls.RESULT_METRIC_SEPARATOR)
        return result, metric, transform

    @classmethod
    def get_data(cls, trial, target, aggregate=True):
        result, metric, transform = cls.split_target(target)
        if aggregate:
            data = trial.get_stats(result, transform=transform)
            return data[metric] if metric in data.columns else data.loc[metric]
        else:
            data = trial.get_data(result, transform=transform)
            return [df[metric] for df in data]

    def __init__(self, logdir, name=None):
        self.logdir = logdir
        self.name = name

        self._trials = {}
        self._params = defaultdict(set)

    @property
    def params(self):
        return {key: sorted(values) for key, values in self._params.items()}

    @property
    def results(self):
        return sorted(self[0][0])

    @property
    def metrics(self):
        return self[0].metrics

    @property
    def size(self):
        n_trials = len(self)
        total_runs = sum(len(trial) for trial in self._trials.values())
        n_runs_per_trial = total_runs / n_trials
        return (n_trials, n_runs_per_trial)

    def info(self):
        n_trials, runs_per_trial = self.size
        print(f"<{self}>")
        print(f"TrialIndex: {n_trials} trials, {runs_per_trial} runs per trial.")
        print(f"ParamIndex: {len(self.params)} parameters.")
        for param, values in self.params.items():
            print(f"  - {param} = [{', '.join(list(map(str, values)))}]")
        print(f"ResultIndex: {len(self.results)} result files.")
        for result, metrics in self.metrics.items():
            print(f"  - {result}({', '.join(metrics)})")

    def setup(self):
        for dirname, subdirs, filenames in os.walk(self.logdir):
            if Experiment.is_trial_dir(dirname):
                trial = Trial.from_directory(dirname)
                self._trials[dirname] = trial

                for key, value in trial.config.items():
                    if isinstance(value, list):
                        value = tuple(value)
                    self._params[key].add(value)

    def get(self, params_values):
        trials = {}
        for name, trial in self._trials.items():
            if all(pv in name for pv in params_values):
                trials[name] = trial
        return trials

    def __str__(self):
        return f"ExperimentAnalysis(logdir='{self.logdir}', name='{self.name}')"

    def __len__(self):
        return len(self._trials)

    def __getitem__(self, i):
        return list(self._trials.items())[i][1]



import pytest

from tuneconfig.grid import TrialGrid


@pytest.fixture(scope="module")
def grid(analysis_list):
    return TrialGrid(*analysis_list)


def test_grid_select_empty(grid):
    anchors = []
    rslt = grid.select(anchors)
    assert rslt is grid
    assert hasattr(grid, "_trials")
    assert isinstance(grid._trials, dict)
    assert len(grid._trials) == len(grid.analyses)


def test_grid_select_single_anchor(grid):
    anchors = ["batch_size=32"]
    rslt = grid.select(anchors)
    assert rslt is grid
    assert hasattr(grid, "_trials")
    assert isinstance(grid._trials, dict)
    assert len(grid._trials) == len(grid.analyses)
    assert all(len(trials) == 6 for trials in grid._trials.values())


def test_grid_select_multiple_anchors(grid):
    anchors = ["batch_size=32", "optimizer=RMSProp"]
    rslt = grid.select(anchors)
    assert rslt is grid
    assert hasattr(grid, "_trials")
    assert isinstance(grid._trials, dict)
    assert len(grid._trials) == len(grid.analyses)
    assert all(len(trials) == 2 for trials in grid._trials.values())


def test_grid_single_target_no_axes(grid, cfg_single_target_no_axes):
    targets, x_axis, y_axis, anchors = cfg_single_target_no_axes
    trial_grid = grid.select(anchors).build(targets, x_axis, y_axis)
    assert grid.shape == (1, 1)
    assert len(trial_grid[None][None]) == len(grid.analyses)


def test_grid_multiple_targets_no_axes(grid, cfg_multiple_targets_no_axes):
    targets, x_axis, y_axis, anchors = cfg_multiple_targets_no_axes
    trial_grid = grid.select(anchors).build(targets, x_axis, y_axis)
    assert grid.shape == (1, 1)
    assert len(trial_grid[None][None]) == len(grid.analyses)
    assert all(len(metrics[-1]) == len(targets) for metrics in trial_grid[None][None])


def test_grid_single_target_multiple_plots_no_axes(grid, cfg_single_target_multiple_plots_no_axes):
    targets, x_axis, y_axis, anchors = cfg_single_target_multiple_plots_no_axes
    trial_grid = grid.select(anchors).build(targets, x_axis, y_axis)
    assert grid.shape == (1, 1)
    assert len(trial_grid[None][None]) == 3 * len(grid.analyses)


def test_grid_multiple_targets_multiple_plots_no_axes(grid, cfg_multiple_targets_multiple_plots_no_axes):
    targets, x_axis, y_axis, anchors = cfg_multiple_targets_multiple_plots_no_axes
    trial_grid = grid.select(anchors).build(targets, x_axis, y_axis)
    assert grid.shape == (1, 1)
    assert len(trial_grid[None][None]) == 3 * len(grid.analyses)
    assert all(len(metrics[-1]) == len(targets) for metrics in trial_grid[None][None])


def test_grid_single_target_x_axis(grid, cfg_single_target_x_axis):
    targets, x_axis, y_axis, anchors = cfg_single_target_x_axis
    trial_grid = grid.select(anchors).build(targets, x_axis, y_axis)
    assert grid.shape == (1, 3)
    assert all(len(plots) == len(grid.analyses) for plots in trial_grid[None].values())


def test_grid_multiple_targets_x_axis(grid, cfg_multiple_targets_x_axis):
    targets, x_axis, y_axis, anchors = cfg_multiple_targets_x_axis
    trial_grid = grid.select(anchors).build(targets, x_axis, y_axis)
    assert grid.shape == (1, 3)
    assert all(len(plots) == len(grid.analyses) for plots in trial_grid[None].values())
    assert all(
        len(metrics[-1]) == len(targets)
        for plots in trial_grid[None].values()
        for metrics in plots
    )


def test_grid_single_target_y_axis(grid, cfg_single_target_y_axis):
    targets, x_axis, y_axis, anchors = cfg_single_target_y_axis
    trial_grid = grid.select(anchors).build(targets, x_axis, y_axis)
    assert grid.shape == (3, 1)
    assert all(
        len(plots) == len(grid.analyses)
        for y_values in trial_grid.keys()
        for plots in trial_grid[y_values].values()
    )


def test_grid_multiple_targets_y_axis(grid, cfg_multiple_targets_y_axis):
    targets, x_axis, y_axis, anchors = cfg_multiple_targets_y_axis
    trial_grid = grid.select(anchors).build(targets, x_axis, y_axis)
    assert grid.shape == (3, 1)
    assert all(
        len(plots) == len(grid.analyses)
        for y_values in trial_grid.keys()
        for plots in trial_grid[y_values].values()
    )
    assert all(
        len(metrics[-1]) == len(targets)
        for y_values in trial_grid.keys()
        for plots in trial_grid[y_values].values()
        for metrics in plots
    )


def test_grid_single_target_both_axis(grid, cfg_single_target_both_axis):
    targets, x_axis, y_axis, anchors = cfg_single_target_both_axis
    trial_grid = grid.select(anchors).build(targets, x_axis, y_axis)
    assert grid.shape == (2, 3)
    assert all(
        len(plots) == len(grid.analyses)
        for y_values in trial_grid.keys()
        for plots in trial_grid[y_values].values()
    )


def test_grid_multiple_targets_both_axis(grid, cfg_multiple_targets_both_axis):
    targets, x_axis, y_axis, anchors = cfg_multiple_targets_both_axis
    trial_grid = grid.select(anchors).build(targets, x_axis, y_axis)
    assert grid.shape == (2, 3)
    assert all(
        len(plots) == len(grid.analyses)
        for y_values in trial_grid.keys()
        for plots in trial_grid[y_values].values()
    )
    assert all(
        len(metrics[-1]) == len(targets)
        for y_values in trial_grid.keys()
        for plots in trial_grid[y_values].values()
        for metrics in plots
    )



from tuneconfig.experiment import Experiment, ExperimentMode
from tuneconfig.analysis import ExperimentAnalysis


def run_experiment(exec_func,
                   config_factory,
                   logdir,
                   num_samples=1,
                   num_workers=1,
                   mode=ExperimentMode.APPEND,
                   name=None,
                   verbose=True):
    # pylint: disable=too-many-arguments

    experiment = Experiment(config_factory, logdir)
    experiment.start()
    experiment.run(
        exec_func, num_samples, num_workers, mode=mode, verbose=verbose)

    analysis = ExperimentAnalysis(logdir, name=name)
    analysis.setup()
    return analysis



from collections import defaultdict
import os
import re

from tuneconfig.analysis import ExperimentAnalysis


def _get_trial_id(commonconfig, x, y, trial_name):
    param_values = set(commonconfig) | set([x, y])
    param_values = filter(None, param_values)
    trial_id = trial_name
    for param_value in param_values:
        trial_id = trial_id.replace(str(param_value), "")
    trial_id = re.sub(r"/{2,}", "/", trial_id)
    if trial_id.startswith("/"):
        trial_id = trial_id[1:]
    if trial_id.endswith("/"):
        trial_id = trial_id[:-1]
    return trial_id


def _get_commonconfig(plots):
    commonconfig = set(filter(None, plots[0][1].split("/")))
    for _, trial_name, _, _ in plots[1:]:
        commonconfig &= set(trial_name.split("/"))
    return sorted(commonconfig)


def _get_metrics(trial, targets, aggregate):
    return {
        target: ExperimentAnalysis.get_data(trial, target, aggregate)
        for target in targets
    }


class TrialGridCell:
    def __init__(self, x, y, plots):
        self.x = x
        self.y = y

        self.plots = plots
        self.analyses = set()
        self.trials = set()
        self.metrics = set()

        self._idx = {}
        self._build_idx()

    def _build_idx(self):
        prefix = os.path.commonpath([x[0] for x in self.plots])
        commonconfig = _get_commonconfig(self.plots)

        idx = 0
        for analysis_logdir, trial_name, trial, metrics in self.plots:
            analysis_id = analysis_logdir.replace(prefix, "")[1:]
            self.analyses.add(analysis_id)

            trial_id = _get_trial_id(commonconfig, self.x, self.y, trial_name)
            self.trials.add(trial_id)

            for metric in metrics:
                self.metrics.add(metric)
                self._idx[(analysis_id, trial_id, metric)] = idx
                idx += 1

        self.analyses = sorted(self.analyses)
        self.trials = sorted(self.trials)
        self.metric = sorted(self.metrics)

    def __len__(self):
        return len(self.plots) * len(self.metrics)

    def index(self, analysis_id, trial_id, metric):
        return self._idx[(analysis_id, trial_id, metric)]


class TrialGrid:
    def __init__(self, analysis, *analyses):
        self.analyses = [analysis, *analyses]

        self._cell_idx = {}

    def select(self, anchors):
        self._trials = {
            analysis.logdir: analysis.get(anchors)
            for analysis in self.analyses
        }
        return self

    def build(self, targets, x_axis=None, y_axis=None, aggregate=True):
        self.x_axis = x_axis
        self.y_axis = y_axis

        self._grid = defaultdict(lambda: defaultdict(list))

        for analysis_id, trials in self._trials.items():
            for name, trial in trials.items():
                name = name.replace(analysis_id, "")
                metrics = _get_metrics(trial, targets, aggregate)

                x_id = None
                if x_axis:
                    x = trial.config.get(x_axis)
                    if isinstance(x, list): x = tuple(x)
                    x_label = f"{x_axis}={x}"
                    x_id = (x, x_label)

                y_id = None
                if y_axis:
                    y = trial.config.get(y_axis)
                    if isinstance(y, list): y = tuple(y)
                    y_label = f"{y_axis}={y}" if y_axis else None
                    y_id = (y, y_label)

                self._grid[y_id][x_id].append(
                    (analysis_id, name, trial, metrics))

        self._build_cell_index()

        return self._grid

    def _build_cell_index(self):
        for j, y in enumerate(self._grid):
            for i, x in enumerate(self._grid[y]):
                plots = self._grid[y][x]
                self._cell_idx[(y, x)] = TrialGridCell(x, y, plots)

    def traverse(self):
        def sort_grid_key(key):
            if not key:
                return key
            return key[0]

        y_values = sorted(self._grid, key=sort_grid_key)

        for j, y in enumerate(y_values):
            x_values = sorted(self._grid[y], key=sort_grid_key)

            for i, x in enumerate(x_values):
                plots = self._grid[y][x]

                prefix = os.path.commonpath([x[0] for x in plots])
                commonconfig = _get_commonconfig(plots)

                for analysis_logdir, trial_name, trial, metrics in plots:
                    analysis_id = analysis_logdir.replace(prefix, "")[1:]
                    trial_id = _get_trial_id(commonconfig, x, y, trial_name)

                    for metric, df in metrics.items():
                        yield (
                            (j, i, y, x, commonconfig),
                            (analysis_id, trial_id, metric),
                            df,
                        )

    def get(self, x, y):
        return self._cell_idx[(y, x)]

    @property
    def shape(self):
        y_values = self._grid.keys()

        x_values = set()
        for values in self._grid.values():
            x_values.update(set(values.keys()))

        return (len(y_values), len(x_values))



import pandas as pd
import pytest


def test_info(trial):
    trial.info()


def test_describe(trial):
    trial.describe()


def test_get_all_data(trial):
    data = trial.get_all_data()
    assert isinstance(data, dict)
    assert len(data) == len(trial.metrics)

    for result, values in data.items():
        assert result in trial.results
        assert isinstance(values, list)
        assert len(values) == len(trial)
        for value in values:
            assert isinstance(value, pd.DataFrame)


def test_get_all_stats(trial):
    stats = trial.get_all_stats()
    assert isinstance(stats, dict)
    assert len(stats) == len(trial.metrics)

    for result, data in stats.items():
        assert result in trial.results
        assert isinstance(data, pd.DataFrame)
        assert all(col[1] in ["min", "max", "mean", "std"]
                   for col in data.columns)



import os
from setuptools import setup, find_packages


def read(filename):
    filepath = os.path.join(os.path.dirname(__file__), filename)
    file = open(filepath, "r")
    return file.read()


setup(
    name="tuneconfig",
    version="v0.12.0",
    author="Thiago P. Bueno",
    author_email="thiago.pbueno@gmail.com",
    description="Hyperparameter config file generator.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    license="GNU General Public License v3.0",
    keywords=["hyperparameter-tuning", "grid-search-hyperparameter"],
    url="https://github.com/thiagopbueno/tuneconfig",
    packages=find_packages(),
    scripts=[],
    install_requires=["matplotlib", "numpy", "pandas", "tqdm",],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



import os
import json
import re

import matplotlib.pyplot as plt

from tuneconfig.grid import TrialGrid


plt.style.use("seaborn-darkgrid")


class ExperimentPlotter:
    """ExperimentPlotter manages plotting for multiple experiment analyses.

    Args:
        analysis (ExperimentAnalysis): (required) analysis object.
        analyses (List(ExperimentAnalysis)): (optional) list of analysis objects.
    """

    def __init__(self, analysis, *analyses, **kwargs):
        self.analyses = [analysis, *analyses]
        self.grid = TrialGrid(*self.analyses)
        self.kwargs = kwargs

    def plot(self, targets, x_axis=None, y_axis=None, anchors=None, **kwargs):
        fmt_kwargs = self.kwargs.copy()
        fmt_kwargs.update(kwargs)

        aggregate = kwargs.get("aggregate", True)
        trial_grid = self.grid.select(anchors).build(targets, x_axis, y_axis, aggregate)

        nrows, ncols = self.grid.shape
        fig, axes = plt.subplots(
            nrows, ncols,
            squeeze=False,
            figsize=fmt_kwargs.get("figsize"),
            sharex=fmt_kwargs.get("sharex", False),
            sharey=fmt_kwargs.get("sharey", False),
        )

        suptitle = None
        for pos, _, _ in self.grid.traverse():
            commonconfig = pos[-1]
            if not suptitle:
                suptitle = set(commonconfig)
            else:
                suptitle &= set(commonconfig)
        fig.suptitle(", ".join(sorted(suptitle)), fontweight="bold")

        for pos, ids, df in self.grid.traverse():
            j, i, y, x, commonconfig = pos
            analysis_id, trial_id, metric = ids
            ax = axes[j][i]

            xy_values = set()
            if x:
                xy_values.add(x[1])
            if y:
                xy_values.add(y[1])
            ax.set_title(", ".join(sorted(xy_values)), fontweight="bold")

            if j == nrows - 1:
                ax.set_xlabel(
                    fmt_kwargs.get("target_x_axis_label"), fontweight="bold")
            if i == 0:
                ax.set_ylabel(
                    fmt_kwargs.get("target_y_axis_label"), fontweight="bold")

            label, index = self._get_plot_label_idx(x, y, *ids)
            self._plot(ax, df, label, index, **fmt_kwargs)

        return fig

    def _plot(self, ax, df, label, index, **kwargs):
        plot_type = kwargs["plot_type"]
        plot_fn = getattr(self, f"_plot_{plot_type}")
        plot_fn(ax, df, label, index, **kwargs)

    def _plot_line(self, ax, df, label, index, **kwargs):
        mean, std = df["mean"], df["std"]
        lower = mean - std
        upper = mean + std

        xs = range(len(mean))
        ax.plot(xs, mean, label=label)
        ax.fill_between(xs, lower, upper, alpha=0.25)
        ax.set_xscale(kwargs.get("xscale", "linear"))
        ax.set_yscale(kwargs.get("yscale", "linear"))
        ax.grid()
        ax.legend()

    def _plot_bar(self, ax, df, label, index, **kwargs):
        mean, std = df["mean"], df["std"]
        x = index * 0.5
        width = 0.35
        ax.bar([x], mean, width, yerr=std, capsize=10, label=label, alpha=0.45)
        ax.set_xticklabels([])
        ax.legend()

    def _plot_boxplot(self, ax, data, label, index, **kwargs):
        label = label[label.find("@") + 2:]
        ax.boxplot(data, positions=[index], labels=[label],
                   vert=kwargs.get("vertical", True), 
                   showmeans=True, meanline=True, showfliers=False)
        ax.grid()

    def _get_plot_label_idx(self, x, y, analysis_id, trial_id, metric):
        label = "/".join(filter(None, [analysis_id, trial_id]))
        label = re.sub(r"/{2,}", "/", label)
        label = f"{metric} @ {label}"
        index = self.grid.get(x, y).index(analysis_id, trial_id, metric)
        return label, index

    def plot_chart_from_spec(self, config_path, show_fig=True, filename=None):
        with open(config_path, "r") as file:
            config = json.load(file)
        fig = self.plot(**config)
        if filename:
            fig.savefig(filename)
        if show_fig:
            plt.show()
